#ifndef __IMAGE_H__
#define __IMAGE_H__

#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>


namespace spectral
{



//! Generic image template.
template <typename T> 
class image
{
public:
    image(uint32_t width, 
          uint32_t height,
          uint32_t channels = 3,
          const T *buf = NULL)
    : m_buffer(NULL)
    , m_width(width)
    , m_height(height)
    , m_channels(channels)
    {
        size_t size = width * height * channels;

        if(size)
        {
            m_buffer = new T[size];

            if(buf && m_buffer)
            {
                for(size_t i=0;i<size;i++)
                {
                     // Inefficient with primitive types.
                     // Should use memcpy
                     m_buffer[i] = buf[i];
                }
            }
            else if(m_buffer)
            {
                for(size_t i=0;i<size;i++)
                {
                    m_buffer[i] = 0;
                }
            } 
        }
    }

    image(const image &other)
        : m_buffer(NULL)
        , m_width(other.get_width())
        , m_height(other.get_height())
        , m_channels(other.get_channels())
    {
        size_t size = m_width * m_height * m_channels;

        const T* buf = other.get_buffer();

        if(size && other.get_buffer())
        {
            m_buffer = new T[size];

            if(buf && m_buffer)
            {
                for(size_t i=0;i<size;i++)
                {
                     // Inefficient with primitive types.
                     // Should use memcpy
                     m_buffer[i] = buf[i];
                }
            }
            else if(m_buffer)
            {
                for(size_t i=0;i<size;i++)
                {
                    m_buffer[i] = 0;
                }
            }
        }
    }

    virtual ~image()
    {
        if(m_buffer)
        {
            delete [] m_buffer;
        }
    }

    //! Get the index of a pixel.
    size_t get_index(uint32_t x, uint32_t y) const
    {
        if((x < m_width) && (y < m_height))
        {
            size_t idx = x + (y * m_width);
            return (size_t)(idx * m_channels);
        }
        return (size_t)(-1);
    }

    //! Get address of a pixel.  This returns NULL if coordinates are invalid.
    const T *get_pixel(uint32_t x, uint32_t y) const
    {
        T *result = (T *)NULL;
        size_t index = get_index(x, y);

        if(m_buffer)
        {
            result = &(m_buffer[index]);
        }

        return result;
    }

    //! Get address of a pixel.  If coordinates are out of range then a point
    //! within the image will be selected according to wrapping rules.
    const T *get_constrained_pixel(int32_t x, int32_t y,
                                   bool wrap_x = false,
                                   bool wrap_y = false) const
    {
        int xx = (int)x;
        int yy = (int)y;

        if(wrap_x)
        {
            xx = constrain_wrap(xx, m_width);
        }
        else
        {
            xx = constrain_reflect(xx, m_width);
        }
        if(wrap_y)
        {
            yy = constrain_wrap(yy, m_height);
        }
        else
        {
            yy = constrain_reflect(yy, m_height);
        }

        return get_pixel(xx, yy);
    }

    //! Get pointer to image buffer.
    T*       get_buffer(void) const {return m_buffer;}

    //! Get image width.
    uint32_t get_width(void)  const {return m_width;}

    //! Get image height.
    uint32_t get_height(void) const {return m_height;}

    //! Get number of channels.
    uint32_t get_channels(void) const {return m_channels;}

    //! Copy data to a pixel, provided that coordinates are valid.
    void set_pixel(uint32_t x, uint32_t y, const T *pixel)
    {
        size_t index = get_index(x, y);
        if(pixel && m_buffer && (index < (m_width * m_height * m_channels)))
        {
            for(size_t i=0;i<(size_t)m_channels;i++)
            {
                m_buffer[i + index] = pixel[i];
            }
        }
    }

    // Expand an image, adding a reflected border.
    image<T> *expand(uint32_t x_border, uint32_t y_border,
                     bool wrap_x, bool wrap_y) const
    {
        image<T> *result(NULL);
    
        uint32_t new_width  = get_width() + (x_border * 2);
        uint32_t new_height = get_height() + (y_border * 2);
        result = new image<T>(new_width, new_height, get_channels());
    
        if(result)
        {
            for (int32_t y=0;uint32_t(y)<new_height;y++)
            {
                int32_t yy;
                yy = int32_t(y) - int32_t(y_border);
                for(int32_t x=0;uint32_t(x)<new_width;x++)
                {
                    int32_t xx;
                    xx = int32_t(x) - int32_t(x_border);
                    result->set_pixel(x,y,get_constrained_pixel(xx,yy,
                                                                wrap_x,wrap_y));
                }
            }
        }
    
        return result;
    }
    
    static int32_t constrain_reflect(int32_t x, uint32_t extent)
    {
        extent--;
    
        if(extent)
        {
            x = abs(x);
    
            while(uint32_t(x) > extent)
            {
                x = abs((2 * (int32_t)extent) - x);
            }
        }
        return x;
    }
    
    static int32_t constrain_wrap(int32_t x, uint32_t extent)
    {
        if(extent)
        {
            while(x < 0)
            {
                x+= extent;
            }
            while(x >= (int32_t)extent)
            {
                x-= extent;
            }
        }
        return x;
    }
    
private:

    T *m_buffer;
    uint32_t m_width, m_height, m_channels;
};

}

#endif

